from .element import Element
from .group import Group
from .node import Node
from .topic import Topic
from .exercise import Exercise
from .task import Task
from . import errors
from . import element

from networkx.algorithms.dag import descendants
from networkx import DiGraph
from networkx.exception import NetworkXError


class Soup:
    def __init__(self, elements):
        self._root_group = Group(
            title='Root', 
            path=(), 
            docname='', 
            userdata=None,
        )

        # build group hierarchy, and throw in remaining elements.
        my_elements = set(elements)
        self._make_hierarchy(self._root_group, my_elements)
        self._add_nodes_to_groups(self._root_group, my_elements)
        assert len(my_elements) == 0, my_elements

        # once the elements have paths in their final hierarchy, we
        # can let them resolve their own stuff. for example, a task
        # initially refers to a person's *path* - final situation
        # should be that a task refers to the Person object directly.
        errs = []
        for element in self._root_group.iter_recursive():
            try:
                element.resolve(self)
            except errors.OpenTrainingError as e:
                errs.append(e)

        if len(errs):
            raise errors.CompoundError(
                'there were errors resolving paths of some elements', errors=errs, 
                # don't know which thing I could refer to when doing a
                # global resolve.
                userdata=None,
            )

        # finally, create graph
        self._worldgraph = self._make_worldgraph(self._root_group)


    @property
    def resolved(self):
        return self._root_group is not None

    @property
    def root(self):
        return self._root_group

    def element_by_path(self, path, userdata):
        return self._root_group.element_by_path(path, userdata=userdata)

    def __iter__(self):
        return self._root_group.iter_recursive()

    def worldgraph(self):
        return self._worldgraph

    def subgraph(self, entrypoints, userdata):
        '''Given entrypoints, compute a subgraph of the world graph that
        contains the entrypoints and all their descendants.

        entrypoints is an iterable of element paths or elements (can
        be mixed)
        '''

        # paranoia
        for e in entrypoints:
            assert isinstance(e, Element)

        topics = set()
        for topic in entrypoints:
            topics.add(topic)
            topics.update(descendants(self._worldgraph, topic))
        return self._worldgraph.subgraph(topics)

    @classmethod
    def _make_hierarchy(cls, root, elements):
        level = 1
        while True:
            all_groups = [g for g in elements if isinstance(g, Group)]
            if not all_groups:   # no more groups
                break
            level_groups = [g for g in all_groups if len(g._requested_path) == level]
            for g in level_groups:
                root.add_element(g, userdata=None)
                elements.remove(g)
            level += 1

    @classmethod
    def _add_nodes_to_groups(cls, root, elements):
        nodes = [n for n in elements if isinstance(n, Element)]
        for n in nodes:
            root.add_element(n, userdata=None)
            elements.remove(n)

    @classmethod
    def _make_worldgraph(cls, root):
        worldgraph = DiGraph()
        collected_errors = []

        for elem in root.iter_recursive(cls=Node):
            worldgraph.add_node(elem)
            for target_path in elem.dependencies:
                try:
                    target_topic = root.element_by_path(target_path, userdata=elem.userdata)
                    worldgraph.add_edge(elem, target_topic)
                except errors.PathNotFound as e:
                    collected_errors.append(
                        errors.DependencyError(
                            f'{elem.docname} ({elem}): dependency {target_path} not found', 
                            userdata=elem))

        if len(collected_errors) != 0:
            raise errors.CompoundError('cannot build world graph', errors=collected_errors, 
                                       # don't know which thing I could refer to when doing a
                                       # global resolve.
                                       userdata=None,
                                       )
        return worldgraph

            
    def _assert_resolved(self):
        if not self.resolved:
            raise errors.NotCommitted('soup not resolved')

    def _assert_unresolved(self):
        if self.resolved:
            raise errors.AlreadyCommitted('soup already resolved')



from .element import Element, verify_is_path


class Node(Element):
    def __init__(self, title, path, docname, dependencies, userdata):
        for d in dependencies:
            verify_is_path(d, userdata=userdata)

        super().__init__(
            title=title, 
            path=path, 
            docname=docname, 
            userdata=userdata)
        self.dependencies = dependencies



from ..core.soup import Soup
from ..core.element import Element
from ..core.topic import Topic
from ..core.exercise import Exercise
from ..core.task import Task
from ..core.group import Group
from ..core.errors import OpenTrainingError, CompoundError
from .errors import SoupAlreadyFailed

from sphinx.util import logging
_logger = logging.getLogger(__name__)


def _prepare_app(app):
    if hasattr(app, '_ot_soup'):
        raise OpenTrainingError('Soup already created, cannot add one more element')
    if hasattr(app, '_ot_soup_failed'):
        raise OpenTrainingError('Soup created already failed once, not retrying')
    if not hasattr(app.env, 'ot_elements'):
        app.env.ot_elements = set()
    app.__class__.soup = create_soup

def sphinx_add_element(app, element):
    _prepare_app(app)
    assert isinstance(element, Element)
    app.env.ot_elements.add(element)    

def sphinx_purge_doc(app, env, docname):
    if hasattr(env, 'ot_elements'):
        env.ot_elements -= {e for e in env.ot_elements if e.docname == docname}

def create_soup(app):
    try:
        return app._ot_soup
    except AttributeError:
        pass

    if hasattr(app, '_ot_soup_failed'):  # do not repeatedly try
        raise SoupAlreadyFailed()

    try:
        app._ot_soup = Soup(app.env.ot_elements)
        return app._ot_soup
    except CompoundError as e:
        app._ot_soup_failed = True
        for err in e:
            _logger.warning(str(err), location=err.userdata)
        raise
    except OpenTrainingError as e:
        app._ot_soup_failed = True
        _logger.warning(str(e), location=e.userdata)
        raise




from . import utils
from . import soup
from .errors import log_and_swallow_error, remove_nodes
from ..core.task import Task
from ..core.errors import OpenTrainingError

from sphinx.util.nodes import set_source_info
from sphinx.util.docutils import SphinxDirective
from docutils import nodes

import itertools
import collections


from sphinx.util import logging
_logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-taskstats', _TaskStatsDirective)
    app.connect('doctree-resolved', _ev_doctree_resolved__expand_taskstats_nodes)

def _ev_doctree_resolved__expand_taskstats_nodes(app, doctree, docname):
    try:
        for n in doctree.traverse(_TaskStatsNode):
            task = app.soup().element_by_path(n.path, userdata=n)

            if not isinstance(task, Task):
                _logger.warning(f'{task} is not a task', location=n)
                continue

            table = nodes.table()
            tgroup = nodes.tgroup(cols=4)
            table += tgroup
            tgroup += nodes.colspec(colwidth=1)
            tgroup += nodes.colspec(colwidth=1)
            tgroup += nodes.colspec(colwidth=1)
            tgroup += nodes.colspec(colwidth=1)

            if 'thead':
                thead = nodes.thead()
                tgroup += thead

                row1 = nodes.row()
                thead += row1
                row2 = nodes.row()
                thead += row2
                row1 += nodes.entry()
                row2 += nodes.entry()

                entry = nodes.entry()
                row1 += entry
                entry += nodes.Text(f'Implementation')
                entry = nodes.entry()
                row2 += entry
                entry += nodes.Text(f'(Points: {task.implementation_points})')

                entry = nodes.entry()
                row1 += entry
                entry += nodes.Text(f'Documentation')
                entry = nodes.entry()
                row2 += entry
                entry += nodes.Text(f'(Points: {task.documentation_points})')

                entry = nodes.entry()
                row1 += entry
                entry += nodes.Text(f'Integration')
                entry = nodes.entry()
                row2 += entry
                entry += nodes.Text(f'(Points: {task.integration_points})')

            if 'tbody':
                tbody = nodes.tbody()
                tgroup += tbody

                worker_shares = collections.defaultdict(lambda: [0,0,0])
                for person, share in task.implementors:
                    worker_shares[person][0] += share
                for person, share in task.documenters:
                    worker_shares[person][1] += share
                for person, share in task.integrators:
                    worker_shares[person][2] += share

                for person in sorted(worker_shares.keys(), key=lambda p: (p.lastname, p.firstname)):
                    row = nodes.row()
                    tbody += row

                    entry = nodes.entry()
                    row += entry
                    p = nodes.paragraph()
                    entry += p
                    if person.firstname and person.lastname:
                        text = f' {person.lastname} {person.firstname}'
                    else:
                        text = person.title
                    p += utils.make_reference(text=text, from_docname=docname, to_docname=person.docname, app=app)

                    # implementation
                    entry = nodes.entry()
                    row += entry
                    share = worker_shares[person][0]
                    if share != 0:
                        entry += nodes.Text(f'{share}%')

                    # documentation
                    entry = nodes.entry()
                    row += entry
                    share = worker_shares[person][1]
                    if share != 0:
                        entry += nodes.Text(f'{share}%')

                    # integration
                    entry = nodes.entry()
                    row += entry
                    share = worker_shares[person][2]
                    if share != 0:
                        entry += nodes.Text(f'{share}%')

            n.replace_self(table)
    except OpenTrainingError as e:
        remove_nodes(doctree, _TaskStatsNode)
        log_and_swallow_error(e, _logger)
 
class _TaskStatsNode(nodes.Element):
    def __init__(self, path):
        super().__init__(self)
        self.path = path

class _TaskStatsDirective(SphinxDirective):
    required_arguments = 1   # path

    def run(self):
        path = utils.element_path(self.arguments[0].strip())

        l = _TaskStatsNode(path=path)
        l.document = self.state.document
        set_source_info(self, l)

        return [l]




from . import topic
from . import exercise
from . import task
from . import taskstats
from . import person
from . import project
from . import project_taskstats
from . import project_personstats
from . import personstats
from . import group
from . import grouplist
from . import graph
from . import soup
from . import dia

def setup(app):
    app.connect('env-purge-doc', soup.sphinx_purge_doc)

    topic.setup(app)
    exercise.setup(app)
    task.setup(app)
    taskstats.setup(app)
    person.setup(app)
    project.setup(app)
    project_taskstats.setup(app)
    project_personstats.setup(app)
    personstats.setup(app)
    group.setup(app)
    grouplist.setup(app)
    graph.setup(app)
    dia.setup(app)



from . import utils
from . import soup
from .errors import log_and_swallow_error, remove_nodes
from ..core.project import Project
from ..core.person import Person
from ..core.task import Task
from ..core.group import Group
from ..core.errors import OpenTrainingError

from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import set_source_info
from docutils import nodes
from docutils.parsers.rst import directives 


from sphinx.util import logging
_logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-project-taskstats', _ProjectTaskStatsDirective)
    app.connect('doctree-resolved', _ev_doctree_resolved__expand_project_taskstats_nodes)

def _ev_doctree_resolved__expand_project_taskstats_nodes(app, doctree, docname):
    try:
        for n in doctree.traverse(_ProjectTaskStatsNode):
            project = app.soup().element_by_path(n.project, userdata=n)

            table = nodes.table()
            tgroup = nodes.tgroup(cols=8)
            table += tgroup
            tgroup += nodes.colspec(colwidth=8)
            tgroup += nodes.colspec(colwidth=4)
            tgroup += nodes.colspec(colwidth=1)
            tgroup += nodes.colspec(colwidth=4)
            tgroup += nodes.colspec(colwidth=1)
            tgroup += nodes.colspec(colwidth=4)
            tgroup += nodes.colspec(colwidth=1)
            tgroup += nodes.colspec(colwidth=2)

            if 'thead':
                thead = nodes.thead()
                tgroup += thead
                row = nodes.row()
                thead += row

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('Task')

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('Implementation Points')

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('%')

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('Documentation Points')

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('%')

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('Integration Points')

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('%')

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('% Total')

            if 'tbody':
                tbody = nodes.tbody()
                tgroup += tbody

                if n.sort_by == 'title': 
                    key=lambda s: s[0].title
                elif n.sort_by == 'percent-total':
                    key=lambda s: s[1][3]
                else:
                    assert False, 'unknown sort_by: '+sort_by

                if n.sort_order == 'ascending':
                    reverse = False
                elif n.sort_order == 'descending':
                    reverse = True
                else:
                    assert False, 'unknown sort_order: '+sort_order

                stats = project.task_stats()
                for task, (implementation_percent, documentation_percent, integration_percent, total_percent) in sorted(stats, key=key, reverse=reverse):
                    row = nodes.row()
                    tbody += row

                    # link to task
                    entry = nodes.entry()
                    row += entry
                    p = nodes.paragraph()
                    entry += p
                    p += [utils.make_reference(text=task.title,
                                               from_docname=docname, to_docname=task.docname,
                                               app=app)]

                    # implementation
                    entry = nodes.entry()
                    row += entry
                    entry += nodes.Text(str(task.implementation_points))

                    entry = nodes.entry()
                    row += entry
                    entry += nodes.Text(str(round(implementation_percent, 2)))

                    # documentation
                    entry = nodes.entry()
                    row += entry
                    entry += nodes.Text(str(task.documentation_points))

                    entry = nodes.entry()
                    row += entry
                    entry += nodes.Text(str(round(documentation_percent, 2)))

                    # integration
                    entry = nodes.entry()
                    row += entry
                    entry += nodes.Text(str(task.integration_points))

                    entry = nodes.entry()
                    row += entry
                    entry += nodes.Text(str(round(integration_percent, 2)))

                    # total
                    entry = nodes.entry()
                    row += entry
                    entry += nodes.Text(str(round(total_percent, 2)))

            n.replace_self([table])
    except OpenTrainingError as e:
        remove_nodes(doctree, _ProjectTaskStatsNode)
        log_and_swallow_error(e, _logger)

class _ProjectTaskStatsNode(nodes.Element):
    def __init__(self, project, sort_by, sort_order):
        super().__init__(self)
        self.project = project
        self.sort_by = sort_by
        self.sort_order = sort_order

class _ProjectTaskStatsDirective(SphinxDirective):
    required_arguments = 1   # path
    option_spec = {
        'sort-by': lambda argument: directives.choice(argument, ('title', 'percent-total')),
        'sort-order': lambda argument: directives.choice(argument, ('ascending', 'descending')),
    }

    def run(self):
        project = utils.element_path(self.arguments[0].strip())

        sort_by = self.options.get('sort-by', 'title')
        sort_order = self.options.get('sort-order', 'ascending')

        tasks = _ProjectTaskStatsNode(
            project = project, 
            sort_by = sort_by, 
            sort_order = sort_order)
        tasks.document = self.state.document
        set_source_info(self, tasks)

        return [tasks]



from .element import Element
from .person import Person
from .task import Task
from .group import Group

from collections import defaultdict
import itertools


class Project(Element):

    def __init__(self, title, path, docname, userdata,                 
                 persons, tasks):
        super().__init__(title=title, path=path, docname=docname, userdata=userdata)
        self.persons = persons
        self.tasks = tasks

    def person_points(self, person):
        '''Given a person, return that person's score as

        :return: ``(implementation_points, documentation_points, integration_points, 
                 implementation_points + documentation_points + integration_points)``
        :rtype: ``(int, int, int, int)``

        '''

        assert self.resolved
        assert type(person) is Person
        assert person in self.persons, (person.path, self.persons)

        implementation_points = documentation_points = integration_points = 0

        for task in self.tasks:
            implementation_points += task.person_implementation_points(person)
            documentation_points += task.person_documentation_points(person)
            integration_points += task.person_integration_points(person)

        return implementation_points, documentation_points, integration_points, \
            implementation_points + documentation_points + integration_points

    def tasks_of_person(self, person):
        '''Given a person, return an iterable over that person's tasks'''

        assert self.resolved
        assert type(person) is Person
        assert person in self.persons

        her_tasks = set()
        for task in self.tasks:
            if person in (p for p,_ in itertools.chain(task.implementors, task.documenters, task.integrators)):
                her_tasks.add(task)
        return her_tasks

    def task_stats(self):
        '''Returns iterable over tuples of the form

        .. code-block:: python

           (task, 
              (implementation_percent,
               documentation_percent, 
               integration_percent, 
               total_percent,
              ),
             )

        Where

        * ``task`` is a :py:class:`task.Task` object``
        * ``implementation_percent``
        * ``documentation_percent``
        * ``integration_percent`` 
        * ``total_percent``

          These represent the completeness, in percent, of
          ``task``. (``total_percent`` is the sum of the other three.)

        See also :py:meth:`task.Task.stats()``

        '''

        for task in self.tasks:
            yield task, task.stats()

    def person_stats(self):
        '''Returns iterable over tuples of the form

        .. code-block:: python

           (person, 
              (implementation_points, 
               documentation_points, 
               integration_points, 
               total_points)
             )

        Where

        * ``person`` is a :py:class:`person.Person` object``
        * ``implementation_points``
        * ``documentation_points``
        * ``integration_points`` 
        * ``total_points``

          These represent the respective scores of ``person`` in the
          project. (``total_points`` is the sum of the other three.)

        '''

        for person in self.persons:
            yield person, self.person_points(person)

    def resolve(self, soup):
        persons = []
        tasks = []
        for person in self.persons:
            if type(person) is Person:
                persons.append(person)
            else:
                elem = soup.element_by_path(person, userdata=self.userdata)
                if isinstance(elem, Person):
                    persons.append(elem)
                elif isinstance(elem, Group):
                    persons.extend(elem.iter_recursive(cls=Person))

        for task in self.tasks:
            if type(task) is Task:
                tasks.append(task)
            else:
                elem = soup.element_by_path(task, userdata=self.userdata)
                if isinstance(elem, Task):
                    tasks.append(elem)
                elif isinstance(elem, Group):
                    tasks.extend(elem.iter_recursive(cls=Task))
        
        self.persons = persons
        self.tasks = tasks

        super().resolve(soup)



class OpenTrainingError(Exception):
    '''Base class for all OpenTraining errors'''
    def __init__(self, msg, userdata):
        super().__init__(msg)
        self.userdata = userdata

class BadPath(OpenTrainingError):
    pass

class DependencyError(OpenTrainingError):
    pass

class PathNotFound(OpenTrainingError):
    pass

class NotCommitted(OpenTrainingError):
    pass

class AlreadyCommitted(OpenTrainingError):
    pass

class CompoundError(OpenTrainingError):
    '''An error that *contains* multiple errors. Used in situations where
    we do not bail out early, but rather continue, and represent the
    user with a list of collected error messages.  '''

    def __init__(self, msg, errors, userdata):
        super().__init__(msg, userdata=userdata)
        self.msg = msg
        self.errors = errors

    def __str__(self):
        msgs = [self.msg]
        for e in self.errors:
            msgs.append(' '*4 + str(e))
        return '\n'.join(msgs)

    def __iter__(self):
        yield CompoundError(self.msg, errors=[], userdata=self.userdata)
        for e in self.errors:
            if isinstance(e, CompoundError):
                yield from e
            else:
                yield e



from ..core.errors import OpenTrainingError, CompoundError

class SoupAlreadyFailed(OpenTrainingError):
    '''A sentinel error for all those places which try to build a
    soup. When that process has failed once, it will fail again, and
    it does not make any sense to retry the whole process.

    '''
    def __init__(self):
        super().__init__('', None)

def remove_nodes(doctree, type):
    for n in doctree.traverse(type):
        n.replace_self([])

def log_and_swallow_error(error, logger):
    if isinstance(error, SoupAlreadyFailed):
        return
    elif isinstance(error, CompoundError):
        for e in error:
            logger.warning(str(e), location=e.userdata)
    elif isinstance(error, OpenTrainingError):
        logger.warning(str(error), location=error.userdata)
    else:
        assert False, 'called with something not from here:' + str(error)



from .node import Node
from . import errors


class Person(Node):  # is-a Node only because that holds path,
                     # docname, and userdata
    def __init__(self, title, path, docname, userdata, 
                 firstname, lastname):
        super().__init__(
            title=title, 
            path=path, 
            docname=docname, 
            userdata=userdata,
            dependencies=[], 
            )

        if not firstname:
            raise OpenTrainingError('firstname must be set')
        if not lastname:
            raise OpenTrainingError('lastname must be set')

        self.firstname = firstname
        self.lastname = lastname



from . import errors
from .element import Element, verify_is_path
from .node import Node


class Group(Element):
    def __init__(self, title, path, docname, userdata):
        super().__init__(
            title=title, 
            path=path, 
            docname=docname, 
            userdata=userdata)
        self._children = {}    # {name: element}

    def __len__(self):
        return len(self._children)

    def add_element(self, element, userdata):
        child_name = element._requested_path[0]
        if len(element._requested_path) == 1: # leaf; add to children
            child = self._children.get(child_name)
            if child:
                raise errors.OpenTrainingError(f'{self}: cannot add "{child_name}"; already exists: {child}', 
                                               userdata=userdata)
            self._children[child_name] = element
            del element._requested_path
            element.parent = self
        else:
            parent = self._children.get(child_name)
            if parent is None:
                raise errors.OpenTrainingError(f'{self}: cannot add "{element._requested_path}": '
                                               f'intermediate "{child_name}" does not exist',
                                               userdata=userdata)
            element._requested_path = element._requested_path[1:]
            parent.add_element(element, userdata=self.userdata)

    def element_by_path(self, path, userdata):
        '''Get element by path. path is relative to this group.'''
        
        verify_is_path(path, userdata)

        element = self._children.get(path[0])
        if element is None:
            raise errors.PathNotFound(f'{self}: no element with name "{path[0]}"', userdata=userdata)
        if len(path) == 1:
            return element

        return element.element_by_path(path[1:], userdata)

    def child_by_name(self, name):
        '''Get direct child element by name.'''
        child = self._children.get(name)
        if child is None:
            raise errors.OpenTrainingError(f'{self}: no child with name {name}')
        return child

    def element_name(self, element):
        '''Get name of element under which it is know to his group.

        Raises OpenTrainingError if element not there.

        '''
        for name, elem in self._children.items():
            if elem is element:
                return name
        else:
            raise errors.OpenTrainingError(f'{element} is not a child of {self}')

    def has_element(self, element):
        '''Does this group have the element?'''
        for name, elem in self._children.items():
            if elem is element:
                return True
        return False

    def iter_recursive(self, cls=None):
        '''Iterator (name, element) over descendants, recursively'''
        if cls:
            assert issubclass(cls, Element)
        else:
            cls = Element

        if issubclass(Group, cls):
            yield self

        for name, elem in self._children.items():
            if isinstance(elem, Group):
                yield from elem.iter_recursive(cls)
            elif isinstance(elem, cls):
                yield elem



from .node import Node
from .element import verify_is_path
from . import errors


class Task(Node):
    IMPLEMENTATION, DOCUMENTATION, INTEGRATION = range(3)

    def __init__(self, 
                 title, path, docname, 
                 dependencies, userdata,

                 implementation_points, implementors,
                 documentation_points, documenters,
                 integration_points, integrators):

        if len(implementors) != 0:
            assert sum(share for person, share in implementors) <= 100
        if len(documenters) != 0:
            assert sum(share for person, share in documenters) <= 100
        if len(integrators) != 0:
            assert sum(share for person, share in integrators) <= 100

        # in tests this is easily gotten wrong
        for p,_ in implementors: verify_is_path(p, userdata=userdata)
        for p,_ in documenters: verify_is_path(p, userdata=userdata)
        for p,_ in integrators: verify_is_path(p, userdata=userdata)

        super().__init__(
            title=title, 
            path=path, 
            docname=docname, 
            # add persons as task dependencies
            dependencies=dependencies + [person for person, share in implementors + documenters + integrators], 
            userdata=userdata)

        self.implementation_points = implementation_points
        self.documentation_points = documentation_points
        self.integration_points = integration_points

        self.implementors = implementors
        self.documenters = documenters
        self.integrators = integrators

    def stats(self):
        implementation_percent = sum(percent for _,percent in self.implementors)
        documentation_percent = sum(percent for _,percent in self.documenters)
        integration_percent = sum(percent for _,percent in self.integrators)

        total_points = sum((self.implementation_points, self.documentation_points, self.integration_points))
        gathered_points = sum((self.implementation_points * implementation_percent/100,
                               self.documentation_points * documentation_percent/100,
                               self.integration_points * integration_percent/100))
        if total_points == 0:
            total_percent = 0.0
        else:
            total_percent = gathered_points / total_points * 100

        return implementation_percent, documentation_percent, integration_percent, total_percent

    def resolve(self, soup):
        errs = []
        resolved_implementors = []
        resolved_documenters = []
        resolved_integrators = []

        for person_path, share in self.implementors:
            try:
                person = soup.element_by_path(person_path, userdata=self.userdata)
                resolved_implementors.append((person, share))
            except errors.OpenTrainingError as e:
                errs.append(e)
        for person_path, share in self.documenters:
            try:
                person = soup.element_by_path(person_path, userdata=self.userdata)
                resolved_documenters.append((person, share))
            except errors.OpenTrainingError as e:
                errs.append(e)
        for person_path, share in self.integrators:
            try:
                person = soup.element_by_path(person_path, userdata=self.userdata)
                resolved_integrators.append((person, share))
            except errors.OpenTrainingError as e:
                errs.append(e)

        if errs:
            raise errors.CompoundError(f'could not resolve some paths of {self}', errs, userdata=None)

        self.implementors = resolved_implementors
        self.documenters = resolved_documenters
        self.integrators = resolved_integrators

        super().resolve(soup)

    def person_implementation_points(self, person):
        return self._person_points(person, self.IMPLEMENTATION)
    def person_documentation_points(self, person):
        return self._person_points(person, self.DOCUMENTATION)
    def person_integration_points(self, person):
        return self._person_points(person, self.INTEGRATION)

    def _person_points(self, person, what):
        assert self.resolved

        if what == self.IMPLEMENTATION:
            workers = self.implementors
            difficulty = self.implementation_points
        elif what == self.DOCUMENTATION:
            workers = self.documenters
            difficulty = self.documentation_points
        elif what == self.INTEGRATION:
            workers = self.integrators
            difficulty = self.integration_points
        else: assert False, what

        score = 0
        for p, share in workers:
            if person is p:
                score += share * difficulty
        return score
        



from . import utils
from . import soup
from .errors import log_and_swallow_error, remove_nodes
from ..core.errors import OpenTrainingError
from ..core.project import Project
from ..core.person import Person
from ..core.task import Task
from ..core.group import Group
from ..core.errors import OpenTrainingError

from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import set_source_info
from docutils import nodes
from docutils.parsers.rst import directives 


from sphinx.util import logging
_logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-personstats', _PersonStatsDirective)
    app.connect('doctree-resolved', _ev_doctree_resolved__expand_personstats_nodes)

def _ev_doctree_resolved__expand_personstats_nodes(app, doctree, docname):
    try:
        for n in doctree.traverse(_PersonStatsNode):
            project = app.soup().element_by_path(n.project, userdata=n)
            person = app.soup().element_by_path(n.person, userdata=n)

            table = nodes.table()
            tgroup = nodes.tgroup(cols=5)
            table += tgroup
            tgroup += nodes.colspec(colwidth=8)
            tgroup += nodes.colspec(colwidth=4)
            tgroup += nodes.colspec(colwidth=4)
            tgroup += nodes.colspec(colwidth=4)
            tgroup += nodes.colspec(colwidth=4)

            if 'thead':
                thead = nodes.thead()
                tgroup += thead
                row = nodes.row()
                thead += row

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('Task')

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('Implementation')

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('Documentation')

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('Integration')

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('Task total')

            if 'tbody':
                tbody = nodes.tbody()
                tgroup += tbody

                for task in project.tasks_of_person(person):
                    row = nodes.row()
                    tbody += row

                    # link to task
                    entry = nodes.entry()
                    row += entry
                    p = nodes.paragraph()
                    entry += p
                    p += utils.make_reference(text=task.title, from_docname=docname, to_docname=task.docname, app=app)

                    # points scored from that task
                    implementation_score = task.person_implementation_points(person)
                    documentation_score = task.person_documentation_points(person)
                    integration_score = task.person_integration_points(person)
                    total_score = implementation_score + documentation_score + integration_score

                    entry = nodes.entry()
                    row += entry
                    entry += nodes.Text(str(implementation_score))

                    entry = nodes.entry()
                    row += entry
                    entry += nodes.Text(str(documentation_score))

                    entry = nodes.entry()
                    row += entry
                    entry += nodes.Text(str(integration_score))

                    entry = nodes.entry()
                    row += entry
                    entry += nodes.Text(str(total_score))

                row = nodes.row()
                tbody += row

                entry = nodes.entry()
                row += entry
                entry += nodes.Text('Total Score')

                entry = nodes.entry()
                row += entry

                entry = nodes.entry()
                row += entry

                entry = nodes.entry()
                row += entry

                entry = nodes.entry()
                row += entry
                *_, total_points = project.person_points(person)
                entry += nodes.Text(str(total_points))

            n.replace_self([table])
    except OpenTrainingError as e:
        remove_nodes(doctree, _PersonStatsNode)
        log_and_swallow_error(e, _logger)

class _PersonStatsNode(nodes.Element):
    def __init__(self, person, project):
        super().__init__(self)
        self.person = person
        self.project = project

class _PersonStatsDirective(SphinxDirective):
    required_arguments = 1   # person
    option_spec = {
        'project': utils.element_path,
    }

    def run(self):
        document = self.state.document

        person = utils.element_path(self.arguments[0].strip())
        project = self.options.get('project')

        if not project:
            return [document.reporter.warning('"project" option missing', line=self.lineno)]

        scores = _PersonStatsNode(
            person = person,
            project = project,
        )

        scores.document = document
        set_source_info(self, scores)

        return [scores]



from . import utils
from . import soup
from .errors import log_and_swallow_error
from ..core.errors import OpenTrainingError
from ..core import errors
from ..core.topic import Topic
from ..core.project import Project
from ..core.errors import OpenTrainingError

from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import set_source_info
from sphinx.util import logging
from docutils import nodes


_logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-project', _ProjectDirective)
    app.connect('doctree-read', _ev_doctree_read__extract_projectnodes)

def _ev_doctree_read__extract_projectnodes(app, doctree):
    try:
        docname = app.env.docname

        for n in doctree.traverse(_ProjectNode):
            soup.sphinx_add_element(app, Project(
                docname = docname, 
                title = utils.get_document_title(docname, doctree), 
                path = n.path,
                userdata = n,
                persons = n.persons,
                tasks = n.tasks,
            ))
            n.replace_self([])
    except OpenTrainingError as e:
        log_and_swallow_error(e, _logger)

class _ProjectNode(nodes.Element):
    def __init__(self, path, persons, tasks):
        super().__init__(self)
        self.path = path
        self.persons = persons
        self.tasks = tasks

class _ProjectDirective(SphinxDirective):
    required_arguments = 1   # path
    option_spec = {
        'persons': utils.list_of_elementpath,
        'tasks': utils.list_of_elementpath,
    }

    def run(self):
        path = utils.element_path(self.arguments[0].strip())
        persons = self.options.get('persons')
        tasks = self.options.get('tasks')

        project = _ProjectNode(
            path = path,
            persons = persons,
            tasks = tasks,
        )

        project.document = self.state.document
        set_source_info(self, project)

        return [project]
